@@ -1,4 +1,5 @@
 use crate::core::{
+    constants::MAX_PIPE_CHUNK_SIZE,
     data::{
         self, DirEntry, FileType, Metadata, Request, RequestPayload, Response, ResponsePayload,
         RunningProcess,
@@ -7,6 +8,7 @@ use crate::core::{
 };
 use log::*;
 use std::{
+    env,
     error::Error,
     net::SocketAddr,
     path::{Path, PathBuf},
@@ -15,7 +17,7 @@ use std::{
     time::SystemTime,
 };
 use tokio::{
-    io::{self, AsyncBufReadExt, AsyncWriteExt, BufReader},
+    io::{self, AsyncReadExt, AsyncWriteExt},
     process::Command,
     sync::{mpsc, oneshot, Mutex},
 };
@@ -50,7 +52,8 @@ pub(super) async fn process(
                 depth,
                 absolute,
                 canonicalize,
-            } => dir_read(path, depth, absolute, canonicalize).await,
+                include_root,
+            } => dir_read(path, depth, absolute, canonicalize, include_root).await,
             RequestPayload::DirCreate { path, all } => dir_create(path, all).await,
             RequestPayload::Remove { path, force } => remove(path, force).await,
             RequestPayload::Copy { src, dst } => copy(src, dst).await,
@@ -60,8 +63,9 @@ pub(super) async fn process(
                 proc_run(tenant.to_string(), addr, state, tx, cmd, args).await
             }
             RequestPayload::ProcKill { id } => proc_kill(state, id).await,
-            RequestPayload::ProcStdin { id, line } => proc_stdin(state, id, line).await,
+            RequestPayload::ProcStdin { id, data } => proc_stdin(state, id, data).await,
             RequestPayload::ProcList {} => proc_list(state).await,
+            RequestPayload::SystemInfo {} => system_info().await,
         }
     }
 
@@ -124,12 +128,14 @@ async fn dir_read(
     depth: usize,
     absolute: bool,
     canonicalize: bool,
+    include_root: bool,
 ) -> Result<ResponsePayload, Box<dyn Error>> {
     // Canonicalize our provided path to ensure that it is exists, not a loop, and absolute
     let root_path = tokio::fs::canonicalize(path).await?;
 
-    // Traverse, but don't include root directory in entries (hence min depth 1)
-    let dir = WalkDir::new(root_path.as_path()).min_depth(1);
+    // Traverse, but don't include root directory in entries (hence min depth 1), unless indicated
+    // to do so (min depth 0)
+    let dir = WalkDir::new(root_path.as_path()).min_depth(if include_root { 0 } else { 1 });
 
     // If depth > 0, will recursively traverse to specified max depth, otherwise
     // performs infinite traversal
@@ -139,9 +145,21 @@ async fn dir_read(
     let mut entries = Vec::new();
     let mut errors = Vec::new();
 
+    #[inline]
+    fn map_file_type(ft: std::fs::FileType) -> FileType {
+        if ft.is_dir() {
+            FileType::Dir
+        } else if ft.is_file() {
+            FileType::File
+        } else {
+            FileType::SymLink
+        }
+    }
+
     for entry in dir {
         match entry.map_err(data::Error::from) {
-            Ok(e) => {
+            // For entries within the root, we want to transform the path based on flags
+            Ok(e) if e.depth() > 0 => {
                 // Canonicalize the path if specified, otherwise just return
                 // the path as is
                 let mut path = if canonicalize {
@@ -170,16 +188,20 @@ async fn dir_read(
 
                 entries.push(DirEntry {
                     path,
-                    file_type: if e.file_type().is_dir() {
-                        FileType::Dir
-                    } else if e.file_type().is_file() {
-                        FileType::File
-                    } else {
-                        FileType::SymLink
-                    },
+                    file_type: map_file_type(e.file_type()),
                     depth: e.depth(),
                 });
             }
+
+            // For the root, we just want to echo back the entry as is
+            Ok(e) => {
+                entries.push(DirEntry {
+                    path: e.path().to_path_buf(),
+                    file_type: map_file_type(e.file_type()),
+                    depth: e.depth(),
+                });
+            }
+
             Err(x) => errors.push(x),
         }
     }
@@ -311,26 +333,33 @@ async fn proc_run(
     // Spawn a task that sends stdout as a response
     let tx_2 = tx.clone();
     let tenant_2 = tenant.clone();
-    let mut stdout = BufReader::new(child.stdout.take().unwrap()).lines();
+    let mut stdout = child.stdout.take().unwrap();
     let stdout_task = tokio::spawn(async move {
+        let mut buf: [u8; MAX_PIPE_CHUNK_SIZE] = [0; MAX_PIPE_CHUNK_SIZE];
         loop {
-            match stdout.next_line().await {
-                Ok(Some(line)) => {
-                    let res = Response::new(
-                        tenant_2.as_str(),
-                        None,
-                        ResponsePayload::ProcStdout { id, line },
-                    );
-                    debug!(
-                        "<Client @ {}> Sending response of type {}",
-                        addr,
-                        res.payload.as_ref()
-                    );
-                    if let Err(_) = tx_2.send(res).await {
+            match stdout.read(&mut buf).await {
+                Ok(n) if n > 0 => match String::from_utf8(buf[..n].to_vec()) {
+                    Ok(data) => {
+                        let res = Response::new(
+                            tenant_2.as_str(),
+                            None,
+                            ResponsePayload::ProcStdout { id, data },
+                        );
+                        debug!(
+                            "<Client @ {}> Sending response of type {}",
+                            addr,
+                            res.payload.as_ref()
+                        );
+                        if let Err(_) = tx_2.send(res).await {
+                            break;
+                        }
+                    }
+                    Err(x) => {
+                        error!("Invalid data read from stdout pipe: {}", x);
                         break;
                     }
-                }
-                Ok(None) => break,
+                },
+                Ok(_) => break,
                 Err(_) => break,
             }
         }
@@ -339,26 +368,33 @@ async fn proc_run(
     // Spawn a task that sends stderr as a response
     let tx_2 = tx.clone();
     let tenant_2 = tenant.clone();
-    let mut stderr = BufReader::new(child.stderr.take().unwrap()).lines();
+    let mut stderr = child.stderr.take().unwrap();
     let stderr_task = tokio::spawn(async move {
+        let mut buf: [u8; MAX_PIPE_CHUNK_SIZE] = [0; MAX_PIPE_CHUNK_SIZE];
         loop {
-            match stderr.next_line().await {
-                Ok(Some(line)) => {
-                    let res = Response::new(
-                        tenant_2.as_str(),
-                        None,
-                        ResponsePayload::ProcStderr { id, line },
-                    );
-                    debug!(
-                        "<Client @ {}> Sending response of type {}",
-                        addr,
-                        res.payload.as_ref()
-                    );
-                    if let Err(_) = tx_2.send(res).await {
+            match stderr.read(&mut buf).await {
+                Ok(n) if n > 0 => match String::from_utf8(buf[..n].to_vec()) {
+                    Ok(data) => {
+                        let res = Response::new(
+                            tenant_2.as_str(),
+                            None,
+                            ResponsePayload::ProcStderr { id, data },
+                        );
+                        debug!(
+                            "<Client @ {}> Sending response of type {}",
+                            addr,
+                            res.payload.as_ref()
+                        );
+                        if let Err(_) = tx_2.send(res).await {
+                            break;
+                        }
+                    }
+                    Err(x) => {
+                        error!("Invalid data read from stdout pipe: {}", x);
                         break;
                     }
-                }
-                Ok(None) => break,
+                },
+                Ok(_) => break,
                 Err(_) => break,
             }
         }
@@ -368,17 +404,7 @@ async fn proc_run(
     let mut stdin = child.stdin.take().unwrap();
     let (stdin_tx, mut stdin_rx) = mpsc::channel::<String>(1);
     tokio::spawn(async move {
-        while let Some(mut line) = stdin_rx.recv().await {
-            // NOTE: If the line coming in does not have a newline character,
-            //       we must add it to properly flush to stdin of process
-            //
-            //       Given that our data structure is called line, we assume
-            //       that it represents a complete line whether or not it
-            //       ends with the linefeed character
-            if !line.ends_with('\n') {
-                line.push('\n');
-            }
-
+        while let Some(line) = stdin_rx.recv().await {
             if let Err(x) = stdin.write_all(line.as_bytes()).await {
                 error!("Failed to send stdin to process {}: {}", id, x);
                 break;
@@ -495,10 +521,10 @@ async fn proc_kill(state: HState, id: usize) -> Result<ResponsePayload, Box<dyn
 async fn proc_stdin(
     state: HState,
     id: usize,
-    line: String,
+    data: String,
 ) -> Result<ResponsePayload, Box<dyn Error>> {
     if let Some(process) = state.lock().await.processes.get(&id) {
-        process.stdin_tx.send(line).await.map_err(|_| {
+        process.stdin_tx.send(data).await.map_err(|_| {
             io::Error::new(io::ErrorKind::BrokenPipe, "Unable to send stdin to process")
         })?;
     }
@@ -521,3 +547,13 @@ async fn proc_list(state: HState) -> Result<ResponsePayload, Box<dyn Error>> {
             .collect(),
     })
 }
+
+async fn system_info() -> Result<ResponsePayload, Box<dyn Error>> {
+    Ok(ResponsePayload::SystemInfo {
+        family: env::consts::FAMILY.to_string(),
+        os: env::consts::OS.to_string(),
+        arch: env::consts::ARCH.to_string(),
+        current_dir: env::current_dir().unwrap_or_default(),
+        main_separator: std::path::MAIN_SEPARATOR,
+    })
+}
