@@ -117,6 +117,14 @@ pub enum RequestPayload {
         /// returned, even if canonicalize is flagged as true
         #[structopt(short, long)]
         canonicalize: bool,
+
+        /// Whether or not to include the root directory in the retrieved
+        /// entries
+        ///
+        /// If included, the root directory will also be a canonicalized,
+        /// absolute path and will not follow any of the other flags
+        #[structopt(long)]
+        include_root: bool,
     },
 
     /// Creates a directory on the remote machine
@@ -184,12 +192,15 @@ pub enum RequestPayload {
         /// Id of the actively-running process to send stdin data
         id: usize,
 
-        /// Complete line to stdin of process
-        line: String,
+        /// Data to send to a process's stdin pipe
+        data: String,
     },
 
     /// Retrieve a list of all processes being managed by the remote server
     ProcList {},
+
+    /// Retrieve information about the server and the system it is on
+    SystemInfo {},
 }
 
 /// Represents an response to a request performed on the remote machine
@@ -286,8 +297,8 @@ pub enum ResponsePayload {
         /// Arbitrary id associated with running process
         id: usize,
 
-        /// Line sent to stdout by the process
-        line: String,
+        /// Data read from a process' stdout pipe
+        data: String,
     },
 
     /// Actively-transmitted stderr as part of running process
@@ -295,8 +306,8 @@ pub enum ResponsePayload {
         /// Arbitrary id associated with running process
         id: usize,
 
-        /// Line sent to stderr by the process
-        line: String,
+        /// Data read from a process' stderr pipe
+        data: String,
     },
 
     /// Response to a process finishing
@@ -316,6 +327,28 @@ pub enum ResponsePayload {
         /// List of managed processes
         entries: Vec<RunningProcess>,
     },
+
+    /// Response to retrieving information about the server and the system it is on
+    SystemInfo {
+        /// Family of the operating system as described in
+        /// https://doc.rust-lang.org/std/env/consts/constant.FAMILY.html
+        family: String,
+
+        /// Name of the specific operating system as described in
+        /// https://doc.rust-lang.org/std/env/consts/constant.OS.html
+        os: String,
+
+        /// Architecture of the CPI as described in
+        /// https://doc.rust-lang.org/std/env/consts/constant.ARCH.html
+        arch: String,
+
+        /// Current working directory of the running server process
+        current_dir: PathBuf,
+
+        /// Primary separator for path components for the current platform
+        /// as defined in https://doc.rust-lang.org/std/path/constant.MAIN_SEPARATOR.html
+        main_separator: char,
+    },
 }
 
 /// Represents information about a single entry within a directory
